import lxml.html
import lxml.etree

from urllib.parse import urljoin

from w3lib.html import safe_url_string

from webster.utils import validators
from webster.net.request import Request


class Parser:
    """
    A class that represents Parser module 
    used to parse Webster.net.Request objects. 
    
    Attributes
    ----------
    Request : object
        Webster.net.Request object.
    
    Methods
    -------
    parse_anchors
        Parses the Webster.net.Request for new anchors.
    
    """
    
    def __init__(self, request: object) -> None:
        """
        Parameters
        ----------
        Request : object
            Webster.net.Request object of the URL to parse data out of.  
        """
        
        #Check for correct instance
        if not isinstance(request, Request):
            raise TypeError(
                "Expected response type of Request, instead got: "
                , type(request))
        else:
            #if type was correct, initialize the class.
            self.request = request
            self.response = self.request.get()
            if self.response is not None:
                self.extractor = lxml.html.fromstring(self.response) 
        
    def parse_anchors(self) -> list:
        """
        Parses anchors from the Webster.net.Request object and returns
        list of adjacent URLs.  
    
        Returns
        -------
        list
            a list of parsed anchors or subURLs found in the website.
    
        """
        if self.response is None:
            raise TypeError("Response body is missing.")
        
        urls = [] 
          
        base_url = self.request.base_url()
        extractor_elements = self.extractor.xpath('.//a/@href')
        
        # find every <a> tag from file, with href attribute.
        for anchor in extractor_elements:
            url = None
            #if anchor is URL instead of relative path add it to the urls list.
            try:
                if validators.URLValidator(anchor):
                    url = anchor
            except (ValueError, TypeError):
                url = urljoin(base_url, anchor)
            
            if url is not None and url not in urls:
                #Escape unsafe characters in the URL according to RFC-3986
                url = safe_url_string(url)
                urls.append(url)   
                    
        return urls
    
    def parse_elements(self, elements: list):
        raise NotImplementedError
    
    def parse_index(self) -> dict:
        """
        Create indices from the site info gathered using
        the Webster.net.Request object.
    
        Returns
        -------
        dict
            dataset of the indices. This dataset can be 
            stored to the database.
        """
        
        root_url = self.request._get_url()
        adjacent = self.parse_anchors()
        title = self.extractor.xpath('//head/title') or None
        description = self.extractor.xpath("//meta[@name='description']/@content") or None
        kwords = self.extractor.xpath("//meta[@name='keywords']/@content") or None
        content = lxml.html.tostring(self.extractor, method='text', encoding='unicode')
        
        indice = {
            "url": root_url,
            "adjacents": adjacent,
            "title": title,
            "description": description,
            "keywords": kwords,
            "text": content
        }
        
        return indice
        



from pymongo import MongoClient

class DatabasePipeline:
    
    collection = "webster"
    
    def __init__(self, db_url: str, db_object: object) -> None:
        self.db_url = db_url
        self.db_object = db_object
    
    @classmethod
    def from_crawler(cls, crawler: object) -> object:
        raise NotImplementedError
        return cls(
            db_url = crawler.settings.MONGO_URL,
            db_object = crawler.settings.MONGO_DB
        )
    
    def open_connection(self) -> None:
        self.client = MongoClient(self.db_url)
        self.connection = self.client[self.db_object]
    
    def close_connection(self) -> None:
        self.client.close()
    
    def feed_database(self, data: dict) -> None:
        self.connection[self.collection].insert_one(data)
        
    


"""
This module contains the default settings.

Group similar settings in logical manner.
"""

from os import path

from datetime import datetime

###########################--SETTINGS--###########################

#===================---PATH SETTINGS---===================
TIME_NOW = datetime.now()
PATH = path

#Path to downloads folder
DL_DIR = "downloads/html/"
PARSE_DIR = "downloads/json/"

#==================---CRAWLER SETTINGS---==================
OBEY_ROBOTSTXT = True

#=================---DATABASE SETTINGS---==================
MONGO_URL = ""
MONGO_DB = ""

###########################-/SETTINGS/-###########################


import re

def ModeValidator(mode: str) -> bool:
    """
    Validates the passed mode type
    
    Parameters
    ----------
    mode : str
        Given mode. Valid modes are: "manual".    

    Raises
    ------
    TypeError
        If mode is not a valid mode.
    
    Returns
    -------
    boolean
        True if mode of accepted type.
        
    """
    
    valid_modes = ["auto", "manual"]
    
    if mode in valid_modes:
        return True
    else: raise TypeError(
                    "Expected Mode type of string, instead got: "
                    , type(mode))

def URLValidator(url: list) -> bool:
    """
    Validates the URL or list of URLs
    
    Parameters
    ----------
    url : str | list
        Given URL or list of URLs

    Raises
    ------
    TypeError
        If URL is not of accepted form or not an valid URL.
    
    Returns
    -------
    bool
        True if URL of accepted form and False if it was not.
    """
    
    def _validate(input: str) -> bool:
        if not isinstance(input, str):
            raise TypeError(
                "Check input list content! Expected URL type of string, instead got: "
                , type(input))
            
        return bool(re.match(
            r"(https?|ftp)://"          # protocol
            r"(\w+(\-\w+)*\.)?"         # host (optional)
            r"((\w+(\-\w+)*)\.(\w+))"   # domain
            r"(\.\w+)*"                 # top-level domain (optional, can have > 1)
            r"([\w\-\._\~/]*)*(?<!\.)"  # path, params, anchors, etc. (optional)
        , input))
        
    if not isinstance(url, str):
        if not isinstance(url, list):
            raise TypeError(
                    "Expected URL type of list or string, instead got: "
                    , type(url))
        else:
            check = all([_validate(x) for x in url])       
    else: check = _validate(url)       
    if check:
        return check
    else:
        raise ValueError(
            "URL not of accepted form, please check form: ", url)  




from urllib import robotparser
from webster.utils import validators


class RobotParser:
    
    def __init__(self, url: str) -> None:
        """
        Parameters
        ----------
        url : str
            Robot.txt URL of current Webster.net.Request object.
            eg. Request.base_url + "robots.txt"
        """
        self.robotparser = robotparser.RobotFileParser()
        self.root_url = ""
        
        if validators.URLValidator(url):
            self.root_url = url
        else: 
            raise ValueError ("Given robots.txt url not valid")
        
        #Set the url for robots.txt to the crawler and read it.
        self.robotparser.set_url(self.root_url)
        self.robotparser.read()
        
    def allowed(self, url: str, user_agent = "*") -> bool:
        """Returns True if the useragent is allowed to fetch the url according 
        to the rules contained in the parsed robots.txt file.

        Parameters
        ----------
        url : str 
            URL to check if its allowed to scrape.
            user_agent (str, optional): Used user agent. Defaults to "*".

        Returns
        -------
        bool
            Boolean value True its fine to crawl. False if not.
        """
        return self.robotparser.can_fetch(user_agent, url)
    
    def delay(self, user_agent = "*") -> None or int:
        """Returns the value of the Crawl-delay parameter from robots.txt 
        for the useragent in question. If there is no such parameter or 
        it doesn’t apply to the useragent specified or the robots.txt 
        entry for this parameter has invalid syntax, return None.

        Parameters
        ----------
        user_agent : (Optional) str 
            Used user agent. Defaults to "*".

        Returns
        -------
        None | int
            Returns the value of the Crawl-delay parameter from robots.txt
            or None if entry for this parameter is invalid.
        """
        
        return self.robotparser.crawl_delay(user_agent)




import pycurl

from io import BytesIO 
from typing import Optional

from webster.utils import url_tools
from webster.utils import validators


class Request:
    
    """
    A class that represents a HTTP request object.
    
    Attributes
    ----------
    url : str
        URL to request data from.
        URL must be in correct form: 
        ex. https://example.com/ or https://www.example.com/
       
    method : str, default = "GET".
        Request method.

    body : (Optional) bytes, default = None.
        Request body. Will be stored as bytes, 
        encoded using utf-8.
        
    encoding : str, default = "utf-8".
        Encoding of request. Encoding is used to encode 
        request body to bytes.
        Also allows for decoding bytes to string.
    
    Methods
    -------
    get
        Returns Webster.net.Request object of HTTP request.
    
    """
    
    def __init__(
        self,
        url: str,
        method: str = "GET",
        encoding: str = "utf-8",
        body: Optional[bytes] = None,
    ) -> None:
        
        self._encoding = encoding
        self.method = str(method).upper()
        self.url = self._set_url(url)
        self.status_code = None
        
        if body is not None:
            if isinstance(body, bytes):
                self.body = body
            else: 
                raise TypeError(
                    "Expected body type of bytes, instead got: "
                    , type(body))
        else: self.body = self.__get()
        
    def _get_url(self) -> str:
        return self.url

    def _set_url(self, url: str) -> str:
        if validators.URLValidator(url):
            return url
        
    def get(self) -> bytes:
        """
        Send GET request to server of the request class object.

        Returns
        -------
        bytes
            Website content.
        """
        return self.body

    def text(self) -> str:
        return self.body.decode(self._encoding)
    
    def base_url(self) -> str:
        """
        Get websites base URL (URL netloc) from the response object

        Returns
        -------
        string
            Base URL (<scheme>://<netloc>/)
        
        """
        return url_tools.base_url(self.url)
    
    def __get(self):
        """
        Helper function for request.get()
        
        This actually is where the magic happens.
        """

        data = None
        b = BytesIO()

        try:
            crl = pycurl.Curl()
            crl.setopt(pycurl.URL, self.url)
            crl.setopt(pycurl.FOLLOWLOCATION, 1)
            crl.setopt(pycurl.CONNECTTIMEOUT, 5)
            crl.setopt(pycurl.TIMEOUT, 8)
            crl.setopt(pycurl.WRITEDATA, b)
            crl.perform()
            
        except (pycurl.error, TypeError):
            #Something went wrong requesting.
            #Could be connection timeout, bad url, no network
            #connection or other stuff
            
            #Return none data. 
            return data
        
        data = b.getvalue()
        
        self.status_code = crl.getinfo(pycurl.HTTP_CODE)
        crl.close()
        
        return data 

    def __str__(self):
        return f"({self.status_code}) <{self.method} : {self.url}>"

    __repr__ = __str__


import os

from webster.conf import settings
from webster.net.request import Request


class Downloader:
    
    """
    A class that represents Downloader module 
    used to download webpages and saving 
    webpages in html format.
    
    Attributes
    ----------
    Request : object
        Webster.net.Request object.
    
    Methods
    -------
    download
        Downloads Webster.net.Request content and saves the content as html file.  
    """
    
    def __init__(self, request: Request) -> None:
        """
        Parameters
        ----------
        Request : object
            Webster.net.Request object of the URL to download.   
        """

        self.filename = None
        self.filepath = None
        
        if not isinstance(request, Request):
            raise TypeError(
                "Expected response type of Request, instead got: "
                , type(request))
        else: 
            self.request = request
              
    def get_request(self):
        return self.request
    
    def get_filename(self):
        return self.filename
    
    def get_filepath(self):
        return self.filepath
    
    def __str__(self):
        return self.filename
          
    __repr__= __str__
       
    def download(self) -> None:
        """
        Download page content as html and store it inside the
        download folder. 
    
        Returns
        -------
        file
            Webster.net.Request contents saved in html format.
    
        """
        self.filename = self.request.url.split(
                "//")[1].replace(
                "/", "") + ".html"   
        self.filepath = settings.PATH.join(
                settings.DL_DIR, self.filename)
            
        try:
            os.makedirs(settings.DL_DIR)
        except FileExistsError:
            pass #Folder already exists not a big deal

        #Try to save the file in DL_DIR folder
        try:       
            if not settings.PATH.isfile(self.filepath):
                with open(self.filepath, 'w') as file:
                    file.write(self.request.url+"\n\n")
                    file.write("File downloaded: ")
                    file.write(settings.TIME_NOW.strftime("%d-%m-%Y, %H:%M:%S")+"\n\n")
                    file.write(self.request.text())

            #If the file already exists, raise error
            else: raise FileExistsError
        except FileExistsError:
            print("File exists already...")                         #SWITCH TO LOGGING
        


import time
import uuid
import queue

from concurrent.futures import ThreadPoolExecutor

from typing import Optional

from webster import robotstxt
from webster.utils import validators
from webster.utils import url_tools
from webster.net.request import Request
from webster.core.parser import Parser
from webster.conf import settings


class Crawler:
    
    """
    A class that represents crawler object used to crawl websites.
    
    Attributes
    ----------
    start_urls : list
        Define starting URLs as a list. First URL in list is then defined
        as the starting point and the rest are stored in Queue.
        URLs must be in correct form: ex. https://example.com/ or https://www.example.com/
       
    allowed_urls : (Optional) list, default = None
        Define allowed URLs to visit.  
         
    mode : (Optional) str, default = auto
        Not Implemented
        
    Methods
    -------
    crawl
        Starts crawler with given starting points.
    
    """
    
    def __init__(self, 
                start_urls: list,
                allowed_urls: Optional[list] = None,
                mode: Optional[list] = None,
        ) -> None:
        
        self._ID = uuid.uuid1()
        
        
        if mode is not None:
            if validators.ModeValidator(mode):
                self.mode = mode
        else: self.mode = "auto"           
        
        if validators.URLValidator(start_urls):
            self.start_urls = start_urls
        else: self.start_urls = None
        
        if allowed_urls in [None, [""], [], [" "]]:
            self.allowed_urls = None
        elif validators.URLValidator(allowed_urls):
            self.allowed_urls = allowed_urls
        
        self.crawling = False
        
        self.queue = queue.Queue()
        
        self.responses = {}
        self.robots_allowed = {}
        self.robots_excluded = {}
    
    def _request(self, url: str) -> Request:
        """
        Helper function for making get requests.
        
        Checks if request is already made to this URL.
        """
        
        if url not in self.responses or url not in self.robots_excluded:
            #BEFORE WE MAKE THE REQUEST TO THE SERVER!!!!
            base_url = url_tools.base_url(url)
            rp = None
            
            allowed = any(base_url
                        in url_tools.base_url(s) 
                        for s 
                        in self.allowed_urls)
            
            if not allowed:
                return None
            
            if base_url not in self.robots_allowed:
                try:
                    rp = robotstxt.RobotParser(base_url + "robots.txt")
                    #Store the RobotParser object to the hashmap for later use cases.
                    self.robots_allowed[base_url] = rp
                    print(f"{self} Robots.txt found {base_url}")           #SWITCH TO LOGGING
                except ValueError:
                    print(f"{self} Robots.txt not found {base_url}")       #SWITCH TO LOGGING
                except Exception as e:
                    print(f"{self} Unexpected exception occured {base_url}:, {e}")       #SWITCH TO LOGGING
            
            if not rp or rp.allowed(url):
                delay = 0
                if settings.OBEY_ROBOTSTXT:
                    if rp is not None and rp.delay():
                        delay = rp.delay()
                
                #Send the request to the server and sleep for the time of delay parameter.
                try:
                    request = Request(url)
                    self.responses[request.url] = request
                except (ValueError, TypeError) as e:
                    print(f"{self} Error requesting {base_url}:, {e}")       #SWITCH TO LOGGING
                    return None
                
                if request.url is None:
                    return None
                
                print(f"{self} Requesting {request}")                   #SWITCH TO LOGGING
                time.sleep(delay)
                
                return request
                
            else:
                #Request did not respect the robots.txt so we skip that
                #and store the excluded url for later use.
                self.robots_excluded[base_url] = 1
                print(f"{self} Robots.txt not allowing {base_url}")       #SWITCH TO LOGGING
    
    def _start_requests(self, urls: list) -> None:
        """
        Start requesting from the given URLs.
        Send Requests to ThreadPool and execute them using threading.
        
        Put Webster.net.Request objects to queue.
        """
        
        if settings.OBEY_ROBOTSTXT:
            for url in urls:
                if url not in self.responses:
                    response = self._request(url)
                    if response is not None:
                        self.queue.put(response)
                    elif self.queue.qsize() == 0:
                        self.crawling = False

        else:
            #Create thread pool executor. Worker count matches our count of awaiting urls.
            with ThreadPoolExecutor(len(urls)) as executor:
                #Add requests to ThreadPool    
                request_futures = executor.map(lambda url : self._request(url), urls)
                #Add Crawler.Requests to queue
                _ = executor.map(self.queue.put, request_futures)
    
    def _crawl(self, rqs: Request) -> None:
        """
        Helper function for crawling.
        
        Parse request for new URLs or anchors.
        Start requesting new URLs.
        """
    
        print(f"{self} Parsing {rqs}")                                      #SWITCH TO LOGGING
        
        try:
            #Parse response anchors with Parser module.
            response_anchors = Parser(rqs).parse_anchors()
            #Store new URLs to this list later on.
            found_urls = []
            
            #If new anchors were found, add them to the list
            if response_anchors:
                for resp in response_anchors:
                    if resp not in self.responses:
                        found_urls.append(resp)
            
            if found_urls: 
                self._start_requests(found_urls)

                
        
        #Skip invalid requests where Request.body is None
        #and thus cannot be parsed.      
        except TypeError:
            print(f"{self} Skipping {rqs}")                                 #SWITCH TO LOGGING
            
    def crawl(self) -> None:
        """
        Start Webster.Crawler process from the starting urls given as
        class attributes. 
        
        If settings.OBEY_ROBOTSTXT is True Crawler will obey websites robots.txt 
        rules. User can disobey these rules by changing this settings value to False.
        Crawler will continue crawling found urls until there is not any to be found.
    
        Returns
        -------
        dict
            Dictionary of crawled websites as Webster.net.Request objects.
        """
        
        if self.crawling:
            raise RuntimeError("Already crawling!")
        self.crawling = True
        
        if self.start_urls is None:
            return None
        
        #Get requests to queue
        self._start_requests(self.start_urls)

             
        while self.crawling:
            print("Queue size:", self.queue.qsize())        #SWITCH TO LOGGING? NECESSARY????
            next_request = self.queue.get()
            
            #Check for bogus requests
            if next_request is not None:
                self._crawl(next_request)
            
            if self.queue.empty():
                self.crawling = False
            
        return self.responses
                              
    def __str__(self):
        return f"Crawler: " + str(self._ID)
        
if __name__ == "__main__":

    #anchors = list(xs.values())[0][1]

    
    url = "https://webscraper.io/test-sites"
    ws = Crawler([url], allowed_urls=["https://webscraper.io/"])                        
    xs = ws.crawl()
    print(xs)


"""
Helper file for URL related stuff

"""
from urllib.parse import urlparse


def netloc_url(url: str) -> str:
    """
    Get websites URL netloc from the URL string.

    Returns
    -------
    string
        netloc URL (<netloc>)
    
    """
    if not isinstance(url, str):
        raise TypeError(
                "Expected URL type of string, instead got: "
                , type(input))
    
    url = urlparse(url)
        
    netloc = '{uri.netloc}'.format(uri=url)
        
    return netloc

def base_url(url: str) -> str:
    """
    Get websites base URL from the URL string.

    Returns
    -------
    string
        Base URL (<scheme>://<netloc>)
    
    """
    if not isinstance(url, str):
        raise TypeError(
                "Expected URL type of string, instead got: "
                , type(input))
    
    url = urlparse(url)
        
    #Get the "base URL" for the relative URLs to work correctly
    #Base URL consist of URL scheme and netloc basically, ignore anything else.
    #Skip the bogus urls where there is scheme or netloc missing.
    if not url[0] or not url[1]:
        return ""
    base_url = '{uri.scheme}://{uri.netloc}/'.format(uri=url)
            
    return base_url


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['webster',
 'webster.conf',
 'webster.core',
 'webster.net',
 'webster.net.request',
 'webster.pipelines',
 'webster.utils']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.9.1,<5.0.0',
 'pycurl>=7.45.1,<8.0.0',
 'pymongo>=4.2.0,<5.0.0',
 'w3lib>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'webster-crawler',
    'version': '0.1.0',
    'description': 'Website crawler-ish thing?',
    'long_description': '# Webster\n\nWebster is an experimental python web scraping framework for my own learning.\n\nWebster can be used to scrape or download webpages by giving a list of urls to visit and\noptionally a list of urls to stay within.\n\nWebster could be used as a broad crawler also, but it is not optimized for that as of now.\n\n## Version\nWebster has been tested using Python 3.10+\n\n## Project documentation\n1. [Quick start guide](docs/project_docs/quickstart.md)\n2. [Architectural model](docs/project_docs/architecture.md)\n3. [Functional specification](docs/project_docs/functional_spec.md)\n4. [Test documentation](docs/project_docs/testing.md)\n\n\n## Installation\nInstall Webster using one off the following methods\n\nFrom requirements.txt\n```bash\npip install -r requirements.txt\n```\n\n\n### Dependencies\nWebster is using multiple external Python packages some of which are:\n+ [lxml](https://lxml.de/index.html) used for processing html documents.\n+ [w3lib](https://w3lib.readthedocs.io/en/latest/) used for URL handling and prosessing.\n+ [pycURL](http://pycurl.io/) used for networking and making requests to a http server.\n+ [pymongo](https://pymongo.readthedocs.io/en/stable/) tools for working with mongoDB using python\n\nUse [pipreqs](https://pypi.org/project/pipreqs/) to make new requirements.txt\nif new dependencies are added in PR force over requirements.txt\nand make sure to include the new one in PR.\n\n```bash\npipreqs --force\n```\n\nUpdate dependencies if necessary\n\n```bash\npip install -U -r requirements.txt\n```\n\n## Testing\nWebster uses [unittest](https://docs.python.org/3/library/unittest.html) for testing.\n\nTo run all the tests use the following command:\n\n```bash\npython -m unittest discover -v\n```\n\nExample: We want to test Parser module\nTo run single module test use the following command:\n\n```bash\npython -m unittest test.test_parser -v\n```\n\n\n### Coverage\nTo run test coverage use the [coverage](https://coverage.readthedocs.io/en/6.5.0/) module\n\n```bash\npython -m coverage run -m unittest\n```\n\nTo show coverage report in terminal\n\n```bash\npython -m coverage report\n```\n\nOutput should look something like this.\n\n```console\nName                              Stmts   Miss  Cover\n-----------------------------------------------------\ntest/__init__.py                      0      0   100%\ntest/test_data/__init__.py           10      0   100%\ntest/test_downloader.py              43      1    98%\ntest/test_parser.py                  54      1    98%\ntest/test_request.py                 49      2    96%\ntest/test_robotstxt.py               23      1    96%\ntest/test_url_tools.py               22      1    95%\ntest/test_validators.py              53      1    98%\nwebster/__init__.py                   0      0   100%\nwebster/conf/__init__.py              0      0   100%\nwebster/conf/settings.py              9      0   100%\nwebster/core/__init__.py              0      0   100%\nwebster/core/downloader.py           36      8    78%\nwebster/core/parser.py               40      1    98%\nwebster/net/__init__.py               0      0   100%\nwebster/net/request/__init__.py      47      2    96%\nwebster/pipelines/__init__.py         0      0   100%\nwebster/robotstxt.py                 15      2    87%\nwebster/utils/__init__.py             0      0   100%\nwebster/utils/url_tools.py           15      0   100%\nwebster/utils/validators.py          16      7    56%\n-----------------------------------------------------\nTOTAL                               432     27    94%\n```\n\nFor a nicer report and more detailed use the html report\n\n```bash\npython -m coverage html\n\nopen htmlcov/index.html   \n```\n\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License\nBSD 3-Clause License\n\nCopyright (c) 2022, Henri Remonen.\nAll rights reserved.\n\nRedistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:\n\n1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.\n\n2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.\n\n3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.\n\nTHIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.',
    'author': 'Henri Remonen',
    'author_email': 'henri@remonen.fi',
    'maintainer': 'Henri Remonen',
    'maintainer_email': 'henri@remonen.fi',
    'url': 'https://github.com/HRemonen/Webster',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
